/***
*
*	Copyright (c) 2005-2006, AntiHax.Net. All rights reserved.
*	
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted. All use, distribution, or modification is prohibited
*   without donating a cash amount. Visit http://antihax.net/AntiHax:Site_support  (ok really it is free =P).
*
*	Author: "Danni" Wilson
*
****/

#include "plugin.h"
#include "cvar_checks.h"
#include <list>
#include <map>
#include "sha1.h"

using namespace std;

eplayer_t *playerList[32];
list<eplayer_t *> playerHistory;


// Helpers
static const char* itoa(int i)
{
   static char buf[32];
   snprintf(buf, sizeof(buf), "%d", i);
   return buf;
}

static const char* ftoa(double i)
{
   static char buf[32];
   snprintf(buf, sizeof(buf), "%f", i);
   return buf;
}
// @@


void DoTick () 
{ 
		processQueryCvars();
}

class QueryCvar 
{
	public:
		QueryCvar(const char *_name, char _type, char _act, const char *cond, const char *v, const char *f1, const char *f2) 
		{
		
			name = _name;
			Fixed = false;
			switch (_act)
			{
				// This is depreciated
				case 'c':
					action = ACT_CRITICAL;
				default:
					action = ACT_NORMAL;
			}
			switch (_type)
			{
				case 'f':
					type = VAL_FLOAT;
					value.floatval = (float)atof(v);
					if (strlen(f1) > 0)
					{
						Fixed = true;
						fix_value1.floatval = (float)atof(f1);
						fix_value2.floatval = (float)atof(f2);
					}
					 			 
					break;
				case 'i':
					type = VAL_INT;
					value.intval = atoi(v);
					if (strlen(f1) > 0)
					{
						Fixed = true;
						fix_value1.intval = atoi(f1);
						fix_value2.intval = atoi(f2);
					}
					break;
				case 's':
					type = VAL_STRING;
					fix_value1.stringval = NULL;
					fix_value2.stringval = NULL;
					value.stringval = strdup(v);
					if (strlen(f1) > 0)
					{
						Fixed = true;
						fix_value1.stringval = strdup(f1);
						
					}
					break;
					default:
					type = VAL_INT;
					value.intval = atoi(v);
					if (strlen(f1) > 0) 
					{
						Fixed = true;
						fix_value1.intval = atoi(f1);
						fix_value2.intval = atoi(f2);
					}
		
					break;
			}

			if		  (strcmp(cond,"==") == 0) {
				condition = COND_EQUALS;
			} else if (strcmp(cond,"!=") == 0) {
				condition = COND_NOT;
			} else if (strcmp(cond,"<=>") == 0) {
				condition = COND_BOUND;
			} else if (strcmp(cond,">=") == 0) {
				condition = COND_GET;
			} else if (strcmp(cond,"<=") == 0) {
				condition = COND_LET;
			} else if (strcmp(cond,">") == 0) {
				condition = COND_GT;
			} else if (strcmp(cond,"<") == 0) {
				condition = COND_LT;
			} else if (strcmp(cond,"0") == 0) {
				condition = COND_NULL;
			}
		}

		~QueryCvar() 
		{
			if (type == VAL_STRING) 
			{
				free(value.stringval);
				if (Fixed)
					free(fix_value1.stringval);
			}	
		}

		friend bool operator < (const QueryCvar& lhs, const QueryCvar& rhs)
		{
			return lhs.name < rhs.name;
		}

		string name;
		value_u value;
		value_u fix_value1;
		value_u fix_value2;
		bool Fixed;
		unsigned char type;
		unsigned char action;
		unsigned int condition;
};


multimap<char*, QueryCvar*, compare>		query_cvars;

void addQueryCvar(const char *name, char type, char action, const char *cond, const char *v, const char *f1, const char *f2)
{
	if (!g_engfuncs.pfnQueryClientCvarValue2)
		return; // Don't run if we don't have the API
	QueryCvar *c = new QueryCvar(name, type, action, cond, v, f1, f2);
	query_cvars.insert(make_pair((char*)c->name.c_str(), c));
}

static char hex_table[17]= "0123456789abcdef";

/* ------------------------------------------------------------------------ */

static int binToHex (char *bin, char *hex, int lng)
{
  int ch;
  int ch2;

  if (lng <= 0) return -1;

  while (lng-- > 0)
  {
    ch= *bin++ & 0x00FF;
    ch2= (ch >> 4) & 0x000F;
    ch &= 0x000F;

    *hex++= hex_table [ch2];
    *hex++= hex_table [ch];
  }
  *hex= 0;

  return 0;
}


const char *configHash()
{
	string config;
	static multimap<char*, QueryCvar*, compare>::iterator it = NULL;
	for (it = query_cvars.begin(); it != query_cvars.end(); it++)
	{
		QueryCvar *ret = (*it).second;
		config.append(ret->name.c_str());
		config.append(itoa(ret->action));
		config.append(itoa(ret->condition));
		if (ret->type == VAL_STRING) 
		{
			if (ret->Fixed && ret->fix_value1.stringval) config.append((char*)ret->fix_value1.stringval);
			if (ret->value.stringval) config.append((char*)ret->value.stringval);
		} 
		else if (ret->type == VAL_FLOAT)
		{
			if (ret->Fixed)
			{
				config.append(ftoa(ret->fix_value1.floatval));
				config.append(ftoa(ret->fix_value2.floatval));
			}
			config.append(ftoa(ret->value.floatval));
		}
		else if (ret->type == VAL_INT)
		{
			if (ret->Fixed)
			{
				config.append(itoa(ret->fix_value1.intval));
				config.append(itoa(ret->fix_value2.intval));
			}
			config.append(itoa(ret->value.intval));
		}
	}
	
	config.append(ftoa(cvar_e_ticrate->value));
	config.append(ftoa(cvar_e_logging->value));

	std::set<const char*, compare>::const_iterator eit = NULL;
	for (eit = consistancyExceptions.begin(); eit != consistancyExceptions.end(); eit++)
	{
		config.append((*eit));
	}

	SHA1 sha1;
	sha1.addBytes(config.c_str(), config.length());
	char hash[41];
	binToHex((char*)sha1.getDigest(), &hash[0],20);
	return &hash[0];
}

static multimap<char*, QueryCvar*, compare>::iterator iter = NULL;

int queryid;

static void processQueryCvars()
{
	int cnt = 0;
	if (!g_engfuncs.pfnQueryClientCvarValue2)

		return; // Don't run if we don't have the API

	if (iter == query_cvars.end() || iter == NULL) 
		iter = query_cvars.begin();

	for (; iter != query_cvars.end(); iter++)
	{
		cnt++;
		for (int x=0;x < gpGlobals->maxClients;x++)
		{
			edict_t* target = INDEXENT(x);
			if (! FNullEnt(target) && IsAlive(target) && validPlayer(target))
			{
				cnt++;
				queryid++;
				QUERY_CLIENT_CVAR_VALUE2(target, (*iter).second->name.c_str(), queryid);
				cvarRequests.insert(queryid);
			}
		}
		if (cnt>1) 
		{
			iter++;
			break;
		}
	}
}

static void doQueryCvar(const edict_t *pEdict, unsigned char action, const char *cvar, const char *value, QueryCvar* ret, value_u fix) 
{
	cl_printf(NULL, HUD_PRINTTALK, "BAD CVAR %s (%s) - %s => %s", STRING(pEdict->v.netname), GETPLAYERAUTHID((edict_t *)pEdict), cvar, value);

	if (ret->Fixed)
	{
		char fixed[1024];
		
		if (ret->type == VAL_STRING)
		{
			snprintf(fixed, 1024,"%s \"%s\"\n", cvar, fix.stringval);
		} 
		else if (ret->type == VAL_FLOAT)
		{
			snprintf(fixed, 1024,"%s \"%f\"\n", cvar, fix.floatval);
		} 
		else if (ret->type == VAL_INT) 
		{
			snprintf(fixed, 1024, "%s \"%d\"\n", cvar, fix.intval);
		}
		CLIENT_COMMAND((edict_t*)pEdict, fixed);	
	}
}

void clearPlayer(edict_t *pEntity)
{
	int pnum = ENTINDEX(pEntity);
	memset(&playerList[pnum],0,sizeof(eplayer_t*));
}

void logCvar(const edict_t *pEdict, const char *cvar, const char *value)
{
	if (!cvar || !value || strlen(value) <=0)
		return;

	if (!playerHistory.size())
		memset(playerList,0,sizeof(eplayer_t*)*32);
	
	int pnum = ENTINDEX(pEdict);
	const char * name = STRING(pEdict->v.netname);
	const char * authid = GETPLAYERAUTHID((edict_t *)pEdict);
	eplayer_t * p = playerList[pnum];
	if (!p || strcmp(p->steamid, authid) != 0)
	{	
		// This is a new or different player, make a new history file and link.
		p = new eplayer_t;
		playerHistory.push_back(p);
		p->name		= strdup(name);
		p->steamid	= strdup(authid);
		playerList[pnum] = p;
	}
	
	std::multimap<char*, char*, compare>::iterator it = p->cvars.find((char*)cvar);
	if ( it != NULL && it != p->cvars.end() ) 
	{
		
		if (strcmp((*it).second, value) != 0)
		{
			if ((*it).second) free((*it).second);
			(*it).second = strdup(value);
		}
	}
	else 
	{
		char* ncvar = strdup(cvar);
		char* nvalue = strdup(value);
		p->cvars.insert(make_pair((char*)ncvar, (char*)nvalue));
	}
}

void gotQueryCvarReturn(const edict_t *pEdict, int requestId, const char *cvar, const char *value) 
{
	QueryCvar *ret = (*query_cvars.find((char*)cvar)).second;
	logCvar(pEdict, cvar, value);

	switch (ret->condition) 
	{

	case COND_EQUALS:
		if (ret->type == VAL_STRING) 
		{
			if (strcmp(value, ret->value.stringval) == 0)
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} 
		else if (ret->type == VAL_FLOAT)
		{
			if (fabs(atof(value)-ret->value.floatval) < 0.0001) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} 
		else if (ret->type == VAL_INT) 
		{
			if (atoi(value) == ret->value.intval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		}
		break;

	case COND_NOT:
		if (ret->type == VAL_STRING)
		{
			if (!(strcmp(value, ret->value.stringval) == 0)) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} 
		else if (ret->type == VAL_FLOAT) 
		{
			float r = atof(value);
			if (fabs(r-ret->value.floatval) > 0.0001)
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} 
		else if (ret->type != VAL_INT) 
		{
			if (atoi(value) != ret->value.intval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		}
		break;

	case COND_GT:
		if (ret->type == VAL_FLOAT) 
		{
			float r = atof(value);
			if (r > ret->value.floatval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} else if (ret->type == VAL_INT) 
		{
			if (atoi(value) > ret->value.intval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		}
		break;

	case COND_LT:
		if (ret->type == VAL_FLOAT) 
		{
			float r = atof(value);
			if (r < ret->value.floatval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} else if (ret->type == VAL_INT) 
		{
			if (atoi(value) < ret->value.intval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		}
		break;

	case COND_GET:
		if (ret->type == VAL_FLOAT) 
		{
			float r = atof(value);
			if (r >= ret->value.floatval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} else if (ret->type == VAL_INT) 
		{
			if (atoi(value) >= ret->value.intval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		}
		break;

	case COND_LET:
		if (ret->type == VAL_FLOAT) 
		{
			float r = atof(value);
			if (r <= ret->value.floatval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		} else if (ret->type == VAL_INT) 
		{
			if (atoi(value) <= ret->value.intval)
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value1);
			}
		}
		break;

	case COND_BOUND:
		if (ret->type == VAL_FLOAT) 
		{
			float r = atof(value);
			if (r < ret->value.floatval || r > ret->fix_value1.floatval) 
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value2);
			}
		} else if (ret->type == VAL_INT) 
		{
			if (atoi(value) < ret->value.intval || atof(value) > ret->fix_value1.intval)
			{
				doQueryCvar(pEdict, ret->action, cvar, value, ret, ret->fix_value2);
			}
		}
		break;
	}
}
